#include "proximityclient.h"

#include <QFile>
#include <QBuffer>
#include <QTimer>

/*!
  Dieser Konstruktor bereitet einen Socket vor (\sa prepareSocket).
  */
ProximityClient::ProximityClient(QString server, quint16 port, PTNP::msgType msgType)
    : _msgType(msgType), _server(server), _port(port)
{
    this->_socket = NULL;
    this->prepareSocket();
}

/*!
  Der Socket wird geschlossen und gelöscht.
  */
ProximityClient::~ProximityClient()
{
    if(_socket->isOpen())
        _socket->close();
    delete _socket;
}

/*!
  Hier wird die evtl. abgebrockene Verbindung neu aufgebaut
  */
void ProximityClient::reconnect() {
    if (this->_socket != NULL) {
        this->_socket->close();
        this->_socket->deleteLater();
    }
    // eventuelles Speicherloch (weil deleteLater an einem NULL-Objekt ausgeführt wird)
    this->_socket = NULL;
    QTimer::singleShot(1000, this, SLOT(prepareSocket()));
}

/*!
  Diese Methode bereutet einen Socket für die Verbindung zum Server vor.
  */
void ProximityClient::prepareSocket() {
    if (this->_socket == NULL) {
        _socket = new QTcpSocket();
        connect(_socket, SIGNAL(readyRead()), this, SLOT(handleResponse()), Qt::QueuedConnection);
        connect(_socket, SIGNAL(error(QAbstractSocket::SocketError)),
                this, SLOT(handleSocketError(QAbstractSocket::SocketError)));
        connect(_socket, SIGNAL(connected()), this, SLOT(initConnection()));
#ifdef _DEV_
        QTextStream cout(stdout, QIODevice::WriteOnly);
        cout<<"ProximityClient.cpp: connecting to server:"<<this->_server<<" on Port: "<<this->_port<<"... "<<flush;
#endif
        _socket->connectToHost(this->_server,this->_port,QIODevice::ReadWrite);
#ifdef _DEV_
            cout<<"connected\n"<<flush;
#endif
    }
}

/*!
  Diese Methode sendet nach dem erfolgreichen Herstelen einer Verbindung die Controllmessage, welche Nachrichten er bekommen möchte.
  Die Methode sollte für die einzelnen Clients überschrieben werden - hier wird eine "gib mir alle Nachrichten" Nachricht abgeschickt.
  */
void ProximityClient::initConnection()
{
#ifdef _DEV_
    QTextStream cout(stdout, QIODevice::WriteOnly);
    cout<<"ProximityClient.cpp: Client: versuche, zu schreiben\n"<<flush;
#endif
    QByteArray block;
    QDataStream out(&block, QIODevice::WriteOnly);

    out.setVersion(QDataStream::Qt_4_4);
    out.setByteOrder(QDataStream::BigEndian);

    out<<_msgType;
    _socket->write(block);
#ifdef _DEV_
    cout<<"ProximityClient.cpp: erfolg!!!\n"<<flush;
#endif
}

/*!
  Diese Methode wird sinnvoll, wenn eine Art Authentifizierung ins
  Protokoll eingearbeitet wird.
  Momentan nicht verwendet!
  */
void ProximityClient::handleInitResponse()
{
#ifdef _DEV_
    QTextStream cout(stdout, QIODevice::WriteOnly);
    cout<<"ProximityClient.cpp: hab' das erste Paket bekommen\n"<<flush;
#endif
    disconnect(_socket, SIGNAL(readyRead()), this, SLOT(handleInitResponse()));
    connect(_socket, SIGNAL(readyRead()), this, SLOT(handleResponse()));
    _socket->readAll();
}

/*!
  Diese Methode nimmt vom Server Pakete entgegen und
  wertet diese aus - this is where the magic happens!
  */
void ProximityClient::handleResponse()
{
#ifdef _DEV_
    QTextStream cout(stdout, QIODevice::WriteOnly);
#endif
    QTextStream cerr(stderr, QIODevice::WriteOnly);
    QDataStream in(_socket);
    QByteArray data;
    QBuffer buffer;
    quint8 header;

    in.setVersion(QDataStream::Qt_4_4);
    in.setByteOrder(QDataStream::BigEndian);
    if (_socket != NULL) {
        while((_socket->bytesAvailable()))
        {
            in.readRawData((char*)&header, 1);
            switch(header) {
            case PTNP::ALL_ANTENNAS_FIXED_TAGS:
                handleAllAntennasPackage();
                break;
            case PTNP::BEACON_MSG:
                handleBeaconPackage();
                break;
            case PTNP::BUTTON_MSG:
                handleButtonPackage();
                break;
            case PTNP::TAG_TO_TAG_CONTACT_NEWS:
                handleTagToTagContactNewsPackage();
                break;
            case PTNP::TAG_TO_TAG_CONTACT_RUNNING:
                handleTagToTagContactRunningPackage();
                break;
            case PTNP::POSITION_MSG:
                handleTagPositionMsg();
                break;
            default:
                _socket->readAll();
                cerr<<"ProximityClient.cpp: unbekannter Pakettyp!\n"<<flush;
            }
        }
    }
}

/*!
  Diese Methode behandelt sämtliche Fehler, die am Socket auftreten können.
  \param e der Socketfehler
  */
void ProximityClient::handleSocketError(QAbstractSocket::SocketError e)
{
    QTextStream cerr(stderr, QIODevice::WriteOnly);
    cerr << "ProximityClient.cpp: ERROR: ";
    switch(e) {
        case QAbstractSocket::ConnectionRefusedError:
            cerr << "ProximityClient.cpp: connection refused!\n";
            break;
        case QAbstractSocket::RemoteHostClosedError:
            cerr << "ProximityClient.cpp: the remote host closed the connection!\n";
            break;
        case QAbstractSocket::HostNotFoundError:
            cerr << "ProximityClient.cpp: host not found!\n";
            break;
        case QAbstractSocket::SocketAccessError:
            cerr << "ProximityClient.cpp: socket operation failed! Not enough privileges?\n";
            break;
        case QAbstractSocket::SocketResourceError:
            cerr << "ProximityClient.cpp: local system ran out of resources!\n";
            break;
        case QAbstractSocket::SocketTimeoutError:
            cerr << "ProximityClient.cpp: the socket operation timed out!\n";
            break;
        case QAbstractSocket::DatagramTooLargeError:
            cerr << "ProximityClient.cpp: datagram was larger than the operating system's limit!\n";
            break;
        case QAbstractSocket::NetworkError:
            cerr << "ProximityClient.cpp: an error occured with the network!\n";
            break;
        case QAbstractSocket::AddressInUseError:
            cerr << "ProximityClient.cpp: the address specified is already in use and was set to be exclusive!\n";
            break;
        case QAbstractSocket::SocketAddressNotAvailableError:
            cerr << "ProximityClient.cpp: the address specified does not belong to the host!\n";
            break;
        case QAbstractSocket::UnsupportedSocketOperationError:
            cerr << "ProximityClient.cpp: the requested socket operation is not supportet by the local operating system!\n";
            break;
//        case QAbstractSocket::ProxyAuthenticationRequiredError:
//            cerr << "the socket is using a proxy, and the proxy requires authentication!\n";
//            break;
//        case QAbstractSocket::SslHandshakeFailedError:
//            cerr << "the SSL/TLS handshake failed, so the connection was closed!\n";
//            break;
//        case QAbstractSocket::UnfinishedSocketOperationError:
//            break;
//        case QAbstractSocket::ProxyConnectionRefusedError:
//            cerr << "could not contact the proxy server because the connection to that server was denied!\n";
//            break;
//        case QAbstractSocket::ProxyConnectionClosedError:
//            cerr << "the connection to the proxy server was closed unexpectedly!\n";
//            break;
//        case QAbstractSocket::ProxyConnectionTimeoutError:
//            cerr << "the connection to the proxy server timed out or the proxy server stopped responding in the authentication phase.\n";
//            break;
//        case QAbstractSocket::ProxyNotFoundError:
//            cerr << "the proxy address was not found!\n";
//            break;
//        case QAbstractSocket::ProxyProtocolError:
//            cerr << "the connection negotiation with the proxy server because the response from the proxy server could not be understood!\n";
//            break;
        case QAbstractSocket::UnknownSocketError:
            cerr << "ProximityClient.cpp: an unidentified error occured.\n";
            break;
        default:
            cerr << "ProximityClient.cpp: something somewhere went terribly wrong!\n";
    }
    cerr << flush;
    emit socketError(e);
    this->reconnect();
}

/*!
  Erstellt aus den Bytes vom Server wieder ein AllAntennasPackage
  emitiert \sa allAntennas .
  */
void ProximityClient::handleAllAntennasPackage()
{
    QDataStream in(_socket);
    quint8 numAntennas;
    quint8 numFixedTags;
    in>>numAntennas;
    in>>numFixedTags;
    QList<PTNP::antenna> antennas;
    QList<PTNP::oid> fixedTags;
    for(int i=0; i<numAntennas; i++) {
        PTNP::antenna a;
        in>>a.id;
        in>>a.flag;
        antennas.append(a);
    }
    for(int i=0; i<numFixedTags; i++) {
        PTNP::oid o;
        in>>o;
        fixedTags.append(o);
    }
    emit allAntennas(antennas, fixedTags);
}

/*!
  Erstellt aus den Bytes vom Server wieder ein TagPositionPaket.
  emitiert \sa tagPositionMsg .
  */
void ProximityClient::handleTagPositionMsg()
{
    QDataStream in(_socket);
    PTNP::tagPositionMsg msg;
    in>>msg.tag1;
    in>>msg.pos.x;
    in>>msg.pos.y;
    in>>msg.pos.f;
    emit tagPositionMsg(msg);
}

/*!
  Erstellt aus den Bytes vom Server wieder ein BeaconPackage.
  emitiert \sa beaconMsg
  */
void ProximityClient::handleBeaconPackage()
{
    QDataStream in(_socket);
    PTNP::beaconMsg package;
    in>>package.tag1;
    in>>package.ant;
    in>>package.strength;
    in>>package.seq;
    emit beaconMsg(package);
}

/*!
  Erstellt aus den Bytes vom Server wieder ein ButtonPackage.
  emitiert \sa buttonMsg
  */
void ProximityClient::handleButtonPackage()
{
    QDataStream in(_socket);
    PTNP::buttonMsg package;
    in>>package.tag1;
    in>>package.dur;
    emit buttonMsg(package);
}

/*!
  Erstellt aus den Bytes vom Server wieder ein TagToTagContactNewsPackage.
  emitiert \sa tagToTagContactNews
  */
void ProximityClient::handleTagToTagContactNewsPackage()
{
    QDataStream in(_socket);
    PTNP::tagToTagContactNews package;
    in>>package.tag1;
    in>>package.tag2;
    in>>package.dur;
    in>>package.end;
    in>>package.flag;
    emit tagToTagContactNews(package);
}

/*!
  Erstellt aus den Bytes vom Server wieder ein TagToTagContactRunningPackage.
  emitiert \sa tagToTagContactRunning
  */
void ProximityClient::handleTagToTagContactRunningPackage()
{
    QDataStream in(_socket);
    PTNP::tagToTagContactRunning package;
    in>>package.tag1;
    in>>package.tag2;
    in>>package.ant;
    in>>package.strength;
    emit tagToTagContactRunning(package);
}

